/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/unTree/delegates/unWindowTreeNodeDelegate.ctl"
#uses "classes/unTree/delegates/unTrendingTreeNodeDelegate.ctl"
#uses "classes/unTree/unTreeProcessContext.ctl"

class UnTreeDelegateFactory{
    public static string getTreeTypeDP(const string treeType, const string widgetType, const bool showAll = false){
        //"fwTT_" skipped intentionally
        if(widgetType == TREE_WIDGET_CURRENT){
            return (showAll ? "*:" : "") + treeType + "Tree";
        } else {
            return (showAll ? "*:" : "") + treeType + widgetType + "Tree";
        }
    }

    public static shared_ptr<FwTreeNodeDelegate> getDelegate(
        const string treeType, const string widgetType
    ){
        if(treeType == TREE_TYPE_WINDOW){
            if(widgetType == TREE_WIDGET_OPERATE){
                return new UnWindowTreeOperateNodeDelegate();
            } else if(widgetType == TREE_WIDGET_CURRENT){
                return new UnWindowTreeOperateNodeDelegate();
            } else if(widgetType == TREE_WIDGET_CONFIGURE){
                return new UnWindowTreeConfigureNodeDelegate();
            } else if(widgetType == TREE_WIDGET_PREVIEW){
                return new UnWindowTreePreviewNodeDelegate();
            } else if(widgetType == TREE_WIDGET_FILE){
                return new UnWindowTreeFileNodeDelegate();
            }
        } else if(treeType == TREE_TYPE_TREND){
            if(widgetType == TREE_WIDGET_OPERATE){
                return new UnTrendingConfigureTreeNodeDelegate();
            } else if(widgetType == TREE_WIDGET_CURRENT){
                return new UnTrendingConfigureTreeNodeDelegate();
            } else if(widgetType == TREE_WIDGET_CONFIGURE){
                return new UnTrendingConfigureTreeNodeDelegate();
            } else if(widgetType == TREE_WIDGET_PREVIEW){
                return new UnTrendingTreePreviewNodeDelegate();
            } else if(widgetType == TREE_WIDGET_FILE){
                return new UnTrendingTreeFileNodeDelegate();
            }
        }
        FwException::raise("Could not create delegate for parameters: " + treeType + ", " + widgetType);
    }
};
