/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/fwTrending/FwTrendingTreeNodeDelegate.ctl"
#uses "classes/fwTree/FwTree_Repository.ctl"
#uses "classes/fwTree/FwTreeNodeDelegate.ctl"
#uses "classes/unTree/delegates/unWindowTreeNodeDelegate.ctl"
#uses "classes/unTree/unTrending_TrendContainer.ctl"
#uses "unTrendTree.ctl"

class UnTrendingOperateTreeNodeDelegate : FwTrendingTreeNodeDelegate{

  public UnTrendingOperateTreeNodeDelegate() : FwTrendingTreeNodeDelegate(FwTrendingTreeMode::navigator){

  }

  public void handleRightClick(FwTreeNodePtr tn){
    if(equalPtr(tn, nullptr))
            return;
    dyn_string navigatorMenu;

    if (tn.getNodeType()!=FwTrending_PlotDp::DPT &&
        tn.getNodeType()!=FwTrending_LegacyPlot::DPT &&
        tn.getNodeType()!=FwTrending_PageDp::DPT &&
        tn.getNodeType()!=FwTrending_LegacyPage::DPT 
        ){
          return;
        }
    

      dynAppend(navigatorMenu,"PUSH_BUTTON, --- "+tn.getName()+" --- , 0, 0");
      dynAppend(navigatorMenu,"SEPARATOR");
      dynAppend(navigatorMenu,"PUSH_BUTTON, Open on top , "+ 1 +", 1");
      
      int answer;
      popupMenu(navigatorMenu, answer);
      switch (answer) {
          case 1:
            openNewTrendInWindow(tn);              
            break;          
          default:
            break;
      }
  }

   public void handleLeftClick(FwTreeNodePtr tn) {
        openNewTrendInHmi(tn);
   }

   public void handleDoubleClick(FwTreeNodePtr tn) {
        //do nothing for double clicking as the WinCC OA tree
        //treats the single click AND the double click event when double clicking...
   }

    private void openNewTrendInHmi( FwTreeNodePtr tn){
      if (equalPtr(tn, nullptr) || tn.isRoot()) {
            return;
        }

        shared_ptr<FwTrending_TrendContainerA> _trendContainer;

        switch (tn.getNodeType()){
          case FwTrending_PlotDp::DPT:
          case FwTrending_LegacyPlot::DPT:
            _trendContainer = new UnTrending_PlotContainer(
              new FwTrending_TrendDataTree(tn));
            break;
          case FwTrending_PageDp::DPT:
          case FwTrending_LegacyPage::DPT:
            _trendContainer = new UnTrending_PageContainer(
              new FwTrending_TrendDataTree(tn));
            break;
          default:
            return;
        }        
        _trendContainer.openInHmi(getNodeName(tn));
    }


    private void openNewTrendInWindow( FwTreeNodePtr tn){
      if (equalPtr(tn, nullptr) || tn.isRoot()) {
            (FwTrending_ActionOpenTrendManager()).executeNoArg();
        } else {
            (FwTrending_ActionDisplayNode(tn, _treeMode)).executeNoArg();
        }
    }


    private string getNodeName(FwTreeNodePtr tn){
      //Need to massage the DP to get the requested nodeName by unTrendTree_handleEventNodeClickOperation()
        string nodeName= tn.getDP();
        nodeName.replace("fwTN_", "");
        return nodeName;
    }

};

class UnTrendingTreePreviewNodeDelegate : UnWindowTreePreviewNodeDelegate {

};

class UnTrendingTreeFileNodeDelegate : UnWindowTreePreviewNodeDelegate {

};

class UnTrendingConfigureTreeNodeDelegate : FwTrendingTreeNodeDelegate{

  public UnTrendingConfigureTreeNodeDelegate() : FwTrendingTreeNodeDelegate(FwTrendingTreeMode::editor){

  }
};
