/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/fwTree/FwTreeNodeDelegate.ctl"
#uses "classes/fwTree/FwTree_Repository.ctl"

class UnWindowTreeNodeDelegate : FwTreeNodeDelegate
{
    protected string getIconAtCurrentDensity(string iconName)
    {
        const mapping densityToSize = makeMapping(0u,"", 1u,"_20", 2u,"_32", 3u,"_40");
        string sIcon="unTree/svg/"+iconName;
        sIcon+=densityToSize.value(getUiDensity(),"");
        sIcon+=".svg";

        if (getPath(PICTURES_REL_PATH,sIcon)=="") return ""; // not found - return empty...
        return sIcon;
    }

    public FwTreeViewItemData getTreeItemData(FwTreeNodePtr tn)
    {
        // start from the parent implementation...
        FwTreeViewItemData data = FwTreeNodeDelegate::getTreeItemData(tn);
        if(tn.name.startsWith("&")) tn.name=substr(tn.name, 5); // cut the &0001

        string tooltip=data.tooltip;
        if (data.icon.isEmpty()) {
            switch (tn.type) {
                case "":
                    data.icon = getIconAtCurrentDensity("folder");
                    break;
                case "_UnPanel":
                    tooltip="Window panel ("+tn._dp+")";
                    data.icon = getIconAtCurrentDensity("panel");
                    if (tn.hasChildren()) {
                        tooltip="Folder with "+tooltip;
                        data.icon=getIconAtCurrentDensity("folder");
                    }
                    break;
            }
        }
        if (tn.isClipboard()) data.icon=getIconAtCurrentDensity("clipboard");

        data.tooltip=tooltip;
        return data;
    }

    public void handleDoubleClick(FwTreeNodePtr tn)
    {
        if (equalPtr(tn, nullptr) || tn.isRoot()) {
            openWindowManager();
        } else {
            displayNode(tn);
        }
    }

    public static void openWindowManager() {
        ChildPanelOnCentral("unTree/unWindowManager.pnl",
                            "Manage Window Panels",
                            makeDynString());
    }

    public void displayNode(FwTreeNodePtr tn, const bool inWindow = true) {
        const string panelDp = tn.getLinkedObj();
        if(!dpExists(panelDp) || dpTypeName(panelDp) != "_UnPanel") return;
        const string panelFile = unPanel_convertDpPanelNameToPanelFileName(panelDp);
        if( getPath(PANELS_REL_PATH,panelFile)=="" || panelFile == "" ) return;
        dyn_string params;
        if(tn.getUserData().count()>0) params = strsplit(tn.getUserData().at(0),";");

        if(inWindow)
            ChildPanelOnCentral(panelFile, tn.getName(), params);
        else
            displayPanelInHmi(tn,panelDp,panelFile, strjoin(params,";"));
    }

    public void handleLeftClick(FwTreeNodePtr tn) {
        if(equalPtr(tn, nullptr))
            return;
        displayNode(tn,false);
    }

    protected void displayPanelInHmi(FwTreeNodePtr tn,const string panelDp,const string panelFile, const string params){
        const bool userDataSmallerThanMaxNav = dynlen(tn.getUserData()) <= c_unPanelConfiguration_maxPanelNavigation;
        const string dpName = userDataSmallerThanMaxNav ? panelDp : tn.getDP();
        if (globalExists("g_unHMI_RunsInNewHMI")) {
            shape mainWindow = getShape(UNHMI_MAIN_MODULE_NAME + "." + UNHMI_MAIN_PANEL_NAME + ":");
            const int currentModule = mainWindow.unHMI_getCurrentModule();
            if ((currentModule >=1) && userDataSmallerThanMaxNav) {
                unHMI_openPreviewModuleHelper(panelFile, currentModule, params, dpName);
            }  else if (currentModule == UNHMI_SPLITTER_MODULE_NUMBER) {
                unHMI_openPreviewModuleHelper(panelFile, currentModule, params, dpName);
            }  else {
                for (int i=1; i<=UNHMI_MAX_NUMBER_OF_PREVIEW_MODULES; i++) {
                    if (mainWindow.unHMI_getIsPreviewModuleOpen(i)){
                    unHMI_openPreviewModuleHelper(panelFile, i, params, dpName);
                    break;
                    }
                }
            }
            return;
        }

        unGraphicalFrame_showUserPanel(panelFile, params, dpName);
        dyn_string exInfo;
        string buttonPanel;
        unGenericDpFunctions_getContextPanel(panelDp, buttonPanel, exInfo);
        if(buttonPanel != "") {
            unGraphicalFrame_showContextualPanel(buttonPanel, dpName);
        }
    }

};

class UnWindowTreeOperateNodeDelegate : UnWindowTreeNodeDelegate
{
    public void handleRightClick(FwTreeNodePtr tn)
    {
        dyn_string navigatorMenu;
        if (equalPtr(tn, nullptr) || tn.isRoot()) {
            dynAppend(navigatorMenu,"PUSH_BUTTON, Manage Window Panels..., "+1+", 1");
        } else {
            int viewEnabled = (tn.getLinkedObj() != "" ? 1:0);
            if (viewEnabled) {
              string panelFile = unPanel_convertDpPanelNameToPanelFileName(tn.getLinkedObj());
              dynAppend(navigatorMenu,"PUSH_BUTTON, "+tn.getName()+ " -> " +panelFile+", "+2+", "+viewEnabled);
              dynAppend(navigatorMenu,"SEPARATOR");
            }
            dynAppend(navigatorMenu,"PUSH_BUTTON, View in HMI, "+2+", "+viewEnabled);
            dynAppend(navigatorMenu,"PUSH_BUTTON, View in separate window, "+3+", "+viewEnabled);
        }
        int answer;
        popupMenu(navigatorMenu, answer);
        switch (answer) {
            case 1:
                openWindowManager();
                break;
            case 2:
                UnWindowTreeNodeDelegate::handleLeftClick(tn);
                break;
            case 3:
                displayNode(tn);
                break;
        }
    }

};

class UnWindowTreeConfigureNodeDelegate : UnWindowTreeNodeDelegate
{
    public void handleRightClick(FwTreeNodePtr tn)
    {
        dyn_string editorMenu;

        if (equalPtr(tn, nullptr)) {
            dynAppend(editorMenu,"PUSH_BUTTON, Manage Window Panels..., "+1+", 1");
        } else {
            int addEnabled = 1, removeEnabled = 1, removeChildrenEnabled = 1, unlinkEnabled = 1, linkEnabled = 1, panelConfigEnabled = 1,
                cutEnabled = 1, copyEnabled = 1, pasteEnabled = 1, renameEnabled = 1, reorderEnabled = 1, manageEnabled = 1;
            if (tn.getClipboard().hasChildren())  {
              cutEnabled = 0;
            } else {
              pasteEnabled = 0;
            }
            if (tn.isRoot()) {
              removeEnabled = cutEnabled = copyEnabled = renameEnabled = linkEnabled = 0;
            }
            if (tn.isClipboard()) {
              removeEnabled = renameEnabled = addEnabled = cutEnabled = copyEnabled = pasteEnabled = linkEnabled = unlinkEnabled = 0;
            }
            if (tn.getLinkedObj()=="" && tn.getNodeType()=="") {
              unlinkEnabled = panelConfigEnabled = 0;
            }
            if (!tn.hasChildren()) {
              removeChildrenEnabled = reorderEnabled = 0;
            }

            FwTreeNodePtr parent=tn.getParent();
            bool parentIsClipboard = (parent!=nullptr && parent.isClipboard());
            if (parentIsClipboard) {
              addEnabled = cutEnabled = copyEnabled = pasteEnabled = 0;
            }

            int linkedObj = (tn.getLinkedObj() != "" ? 1:0);
            if (linkedObj) {
              string panelFile = unPanel_convertDpPanelNameToPanelFileName(tn.getLinkedObj());
              dynAppend(editorMenu,"PUSH_BUTTON, "+tn.getName()+ " -> " +panelFile+", "+15+", "+linkedObj);
              dynAppend(editorMenu,"SEPARATOR");
            }

            dynAppend(editorMenu,"CASCADE_BUTTON, Add, "+addEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Remove, "+5+", "+removeEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Remove Multiple Children, "+6+", "+removeChildrenEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Unlink Panel, "+7+", "+unlinkEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Link Panel, "+8+", "+linkEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Panel Configuration, "+9+", "+panelConfigEnabled);
            dynAppend(editorMenu,"SEPARATOR");
            dynAppend(editorMenu,"PUSH_BUTTON, Cut, "+10+", "+cutEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Copy, "+11+", "+copyEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Paste, "+12+", "+pasteEnabled);
            dynAppend(editorMenu,"SEPARATOR");
            dynAppend(editorMenu,"PUSH_BUTTON, Rename, "+13+", "+renameEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON, Reorder, "+14+", "+reorderEnabled);
            dynAppend(editorMenu,"SEPARATOR");
            dynAppend(editorMenu,"PUSH_BUTTON, Manage Window Panels..., "+1+", "+manageEnabled);

            dynAppend(editorMenu,"Add");
            dynAppend(editorMenu,"PUSH_BUTTON,  Add Node, "+2+", "+addEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON,  Add Panel, "+3+", "+addEnabled);
            dynAppend(editorMenu,"PUSH_BUTTON,  Add Panels, "+4+", "+addEnabled);
        }

        int answer;
        popupMenu(editorMenu, answer);
        switch (answer) {
            case 1:
                openWindowManager();
                break;
            case 2:
                callFunction(FwTreeNodeDelegate::actionAddNode, tn);
                break;
            case 3:
                addPanelNode(tn);
                break;
            case 4:
                addPanelNodes(tn);
                break;
            case 5:
                callFunction(FwTreeNodeDelegate::actionDeleteNode, tn);
                break;
            case 6:
                removeChildNodes(tn);
                break;
            case 7:
                callFunction(FwTreeNodeDelegate::actionUnlinkObject, tn);
                break;
            case 8:
                actionLinkExistingPanel(tn);
                break;
            case 9:
                actionOpenPanelConfiguration(tn);
                break;
            case 10:
                callFunction(FwTreeNodeDelegate::actionCutToClipboard, tn);
                break;
            case 11:
                callFunction(FwTreeNodeDelegate::actionCopyToClipboard, tn);
                break;
            case 12:
                callFunction(FwTreeNodeDelegate::actionPasteFromClipboard, tn);
                break;
            case 13:
                callFunction(FwTreeNodeDelegate::actionRenameNode, tn);
                break;
            case 14:
                callFunction(FwTreeNodeDelegate::actionReorderChildren, tn);
                break;
        }
    }

    public static void addPanelNode(FwTreeNodePtr tn) {
        string selectedPanel, panelDp;
        selectPanelFile(selectedPanel, panelDp);
        if (selectedPanel=="" || panelDp=="") return;

        addPanel(panelDp, tn);
    }

    public static void addPanelNodes(FwTreeNodePtr tn) {
        dyn_string ds;
        dyn_float df;

        ChildPanelOnCentralReturn("unTree/addPanels.pnl", "Add Panels", makeDynString(), df, ds);

        if (dynlen(df)<1 || df[1]==0) return; // panel was cancelled

        for (int i = 1; i <= dynlen(ds); i++) {
          string panelDp;
          string selectedPanel = ds[i];
          dyn_string exceptionInfo;
          unPanel_create(selectedPanel, panelDp, exceptionInfo);
          if (selectedPanel=="" || panelDp=="") continue;

          addPanel(panelDp, tn);
        }
    }

    public static void removeChildNodes(FwTreeNodePtr tn, const bool confirm = true) {
        dyn_string ds;
        dyn_float df;

        FwTreeNodePVec children=tn.getChildren();
        dyn_string dsChildren;

        for (int i=0;i<children.count();i++) {
          dynAppend(dsChildren, children.at(i).getName());
        }

        if(confirm){
            ChildPanelOnCentralReturn("unTree/removePanels.pnl", "Remove Panels", "$sChildren:"+dsChildren, df, ds);

            if (dynlen(df)<1 || df[1]==0) return; // panel was cancelled
        }

        for (int i = 1; i <= dynlen(ds); i++) {
            try {
              string selectedPanel = ds[i];
              string panelDp = unPanel_convertPanelFileNameToDp(selectedPanel);
              if (selectedPanel=="" || panelDp=="") continue;
              dyn_string dsPanelDp = strsplit(panelDp, "-");
              string nodeName = dsPanelDp[dynlen(dsPanelDp)];

              shared_ptr<FwTree_Repository> treeRepo = FwTree_Repository::getInstance();
              FwTreeNodePVec tnToDrop = treeRepo.findBy("name", nodeName);
              for (int i=0; i<tnToDrop.count(); i++) {
                // Remove the proper node in case that there are duplicates or copies in Clipboard
                if (tn == tnToDrop.at(i).getParent()) tnToDrop.at(i).removeMe(true);
              }
            } catch {
              FwException::handleLast();
            }
        }
    }

    public static void actionLinkExistingPanel(FwTreeNodePtr tn)
    {
        string selectedPanel, panelDp;
        selectPanelFile(selectedPanel, panelDp);
        if (panelDp=="" || panelDp==tn.getLinkedObj()) return;
        tn.setProps(dpTypeName(panelDp), panelDp);
    }

    public static void actionOpenPanelConfiguration(FwTreeNodePtr tn)
    {
        ChildPanelOnCentral("vision/unPanel/unPanelConfiguration.pnl",
                            "Panel Configuration",
                            makeDynString("$sDpName:"+tn.getDP()));
    }

    public void handleDrop(FwTreeNodePtr droppedOn, string payload)
    {
        if (payload.isEmpty()) return;

        // payload may be a list of selected plots, pages, etc, separated by newlines, comas or semicolon
        dyn_string payloadItems=strsplit(payload,"\n,;");
        if (payloadItems.isEmpty()) return;

        if (payloadItems.count()>1) handleDropMultiple(droppedOn,payloadItems);

        if (! dpExists(payload)) return;

        switch (dpTypeName(payload)) {
            case "_UnPanel":
              if (droppedOn.isRoot() || droppedOn.isClipboard()) return;
              actionSetLinkedObject(droppedOn, payload, dpTypeName(payload));
              break;
            default:
              DebugTN("Unsupported content dropped on the Trend Tree item: "+payload+" of type "+dpTypeName(payload));
              return;
              break;
        }
    }

    public void handleDropMultiple(FwTreeNodePtr droppedOn, dyn_string payloadItems)
    {
        FwException::raise("Multi-item drag-and-drop not supported");
    }


    private static void addPanel(string panelDp, FwTreeNodePtr tn) {
        try {
          dyn_string dsPanelDp = strsplit(panelDp, "-");
          string nodeName = dsPanelDp[dynlen(dsPanelDp)];
          FwTreeNodePtr newTreeNode = FwTreeNode::create(nodeName, tn, "_UnPanel", panelDp);
        } catch {
          FwException::handleLast();
        }
    }

    private static void selectPanelFile(string &selectedPanel, string &panelDp)
    {
        try {
          fileSelector(selectedPanel, getPath(PANELS_REL_PATH), true, "*.pnl *.xml");
          if((selectedPanel != "")) {
              // remove rel path
              for(int i=1; i<=SEARCH_PATH_LEN; i++) {
                  string path = getPath(PANELS_REL_PATH, "", 1, i);
                  int pos = strpos(selectedPanel, path);
                  if(pos >=0) {
                    selectedPanel = substr(selectedPanel, strlen(path), strlen(selectedPanel) - strlen(path));
                  }
              }
              dyn_string exceptionInfo;
              unPanel_create(selectedPanel, panelDp, exceptionInfo);
          }
        } catch {
          FwException::handleLast();
        }
    }
};

class UnWindowTreePreviewNodeDelegate : UnWindowTreeOperateNodeDelegate {
    // protected mapping _colorMap;
    protected mapping _sourceRepos;

    // public setColorMapping(const mapping map){
    //     _colorMap = map;
    // }

    public void addSourceRepo(const string name, shared_ptr<FwTree_Repository> repo ){
        _sourceRepos[name] = repo;
    }

    public FwTreeNodePtr _decodeNode(const string payload, string& source){
        FwTreeNodePtr result = FwTree_Repository::getDefault().get(payload);
        if(!equalPtr(result,nullptr)){
            source = "default";
            return result;
        }
        for(int i=0; i<_sourceRepos.count(); ++i){
            result = _sourceRepos.valueAt(i).get(payload);
            if(!equalPtr(result,nullptr)){
                source = _sourceRepos.keyAt(i);
                return result;
            }
        }
        source = "unknown";
        return nullptr;

    }

    // public FwTreeViewItemData getTreeItemData(FwTreeNodePtr tn){
    //     FwTreeViewItemData result = UnWindowTreeNodeDelegate::getTreeItemData(tn);
    //     const string id = tn.getDP();
    //     // if(_colorMap.contains(id)){
    //     //     result.hasOptionalProperties = true;
    //     //     // result.bgColor = "red";
    //     //     result.fgColor = _colorMap[id];
    //     //     //result.font = "Arial Black";
    //     // } // else {DebugTN(mappingKeys(_colorMap));};

    //     return result;
    // }

    protected FwTreeNodePtr handleExternalNodeDropedOnItem(FwTreeNodePtr dropOnTN, FwTreeNodePtr droppedTN){ // copy from fwTree.pnl
        int answer;
        dyn_string menu=makeDynString(
            "PUSH_BUTTON,MOVE "+droppedTN.name+",1,0",
            "SEPARATOR",
            "PUSH_BUTTON,Add as child,2,1",
            "PUSH_BUTTON,Insert before,3,"+ (dropOnTN.isClipboard() || dropOnTN.isRoot() ? 0 : 1),
            "PUSH_BUTTON,Append after,4," + (dropOnTN.isClipboard() || dropOnTN.isRoot() ? 0 : 1),
            "SEPARATOR",
            "PUSH_BUTTON,Cancel,99,1"
        );
        popupMenu(menu, answer);

        try {
            if (answer==2) {
                return copyNodeStructure(dropOnTN, droppedTN);
                // droppedTN.move(dropOnTN);
            } else if (answer==3) {
                return handleExternalNodeDroppedBetween(dropOnTN, droppedTN, true);
            } else if (answer==4) {
                return handleExternalNodeDroppedBetween(dropOnTN, droppedTN, false);
            } else {
                return nullptr;
            }
        } catch {
            FwException::handleLast();
        }
    }

    public FwTreeNodePtr handleExternalNodeDroppedBetween(FwTreeNodePtr dropOnTN, FwTreeNodePtr droppedTN, bool dropAbove) {
        FwTreeNodePtr parent = dropOnTN.getParent(); // already checked for null in our caller
        // move of an item... and place it below/above the one we selected
        if (equalPtr(parent, nullptr)) return nullptr; // nothing to do...
        FwTreeNodePtr beforeTN=nullptr; // this means: append
        FwTreeNodePVec siblings=parent.getChildren();
        int idx=siblings.indexOf(dropOnTN);
        if (dropAbove) {
            beforeTN=dropOnTN;
        } else {
            idx++; // advance to the next
            if (idx<siblings.count()) beforeTN=siblings.at(idx); // otherwise it is past the end -> use null to get the append
        }
        FwTreeNodePtr newNode = copyNodeStructure(parent, droppedTN);
        if(beforeTN != nullptr){
            siblings.insertAt(idx,newNode); // vector was created before adding so no need of removing from the end
            parent.reorderChildren(siblings);
        }
        return newNode;
    }
    public void handleDrop(FwTreeNodePtr droppedOn, string payload){
        handleDropFromOtherTree(droppedOn,payload,"OnItem");
    }

    public void handleDropFromOtherTree(FwTreeNodePtr droppedOn, string payload, string position){
        string source;
        FwTreeNodePtr payloadTN = _decodeNode(payload, source);
        FwException::assertNotNull(payloadTN,"Could not find repo for node: "+ payload);
        FwTreeNodePtr insertedNode;
        switch (position) {
            case "OnViewPort": // we may consider a handler of some kind ...
                return;
            case "OnItem":
                insertedNode = handleExternalNodeDropedOnItem(droppedOn, payloadTN);
                break;
            case "AboveItem":
            case "BelowItem":
                insertedNode = handleExternalNodeDroppedBetween(droppedOn, payloadTN, (position=="AboveItem"));
        }
        FwException::assertNotNull(insertedNode,"Could not handle dropped node: "+ payload);
        // if(source == "default"){
        //     _colorMap[insertedNode.getDP()] = "black";
        // } else {
        //     _colorMap[insertedNode.getDP()] = "green";
        // }
    }

    public void handleRightClick(FwTreeNodePtr tn)
    {
        //TODO?
        // dyn_string editorMenu;

        // if (equalPtr(tn, nullptr)) {
        //     dynAppend(editorMenu,"PUSH_BUTTON, Manage Window Panels..., "+1+", 1");
        // } else {
        //     if (tn.isRoot()) {
        //     //   removeEnabled = cutEnabled = copyEnabled = renameEnabled = linkEnabled = 0;
        //     }
        //     // if (tn.isClipboard()) {
        //     //   removeEnabled = renameEnabled = addEnabled = cutEnabled = copyEnabled = pasteEnabled = linkEnabled = unlinkEnabled = 0;
        //     // }

        //     // FwTreeNodePtr parent=tn.getParent();
        //     // bool parentIsClipboard = (parent!=nullptr && parent.isClipboard());
        //     // if (parentIsClipboard) {
        //     //   addEnabled = cutEnabled = copyEnabled = pasteEnabled = 0;
        //     // }

        //     int linkedObj = (tn.getLinkedObj() != "" ? 1:0);
        //     if (linkedObj) {
        //       string panelFile = unPanel_convertDpPanelNameToPanelFileName(tn.getLinkedObj());
        //       dynAppend(editorMenu,"PUSH_BUTTON, "+tn.getName()+ " -> " +panelFile+", "+15+", "+linkedObj);
        //       dynAppend(editorMenu,"SEPARATOR");
        //     }

        //     dynAppend(editorMenu,"PUSH_BUTTON, Add to Export, "+10+", "+copyEnabled);

        //}

        /*int answer;
        popupMenu(editorMenu, answer);
        switch (answer) {
            case 1:
                openWindowManager();
                break;
            case 2:
                callFunction(FwTreeNodeDelegate::actionAddNode, tn);
                break;
            case 3:
                addPanelNode(tn);
                break;
            case 4:
                addPanelNodes(tn);
                break;
            case 5:
                callFunction(FwTreeNodeDelegate::actionDeleteNode, tn);
                break;
            case 6:
                removeChildNodes(tn);
                break;
            case 7:
                callFunction(FwTreeNodeDelegate::actionUnlinkObject, tn);
                break;
            case 8:
                actionLinkExistingPanel(tn);
                break;
            case 9:
                actionOpenPanelConfiguration(tn);
                break;
            case 10:
                callFunction(FwTreeNodeDelegate::actionCutToClipboard, tn);
                break;
            case 11:
                callFunction(FwTreeNodeDelegate::actionCopyToClipboard, tn);
                break;
            case 12:
                callFunction(FwTreeNodeDelegate::actionPasteFromClipboard, tn);
                break;
            case 13:
                callFunction(FwTreeNodeDelegate::actionRenameNode, tn);
                break;
            case 14:
                callFunction(FwTreeNodeDelegate::actionReorderChildren, tn);
                break;
        }*/
    }

};

class UnWindowTreeFileNodeDelegate : UnWindowTreePreviewNodeDelegate {

};
