/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/fwTree/FwTree.ctl"
#uses "classes/unTree/deviceParsers/unTreeDeviceParser.ctl"

class UnTreeDeviceParserPanel : UnTreeDeviceParser {
    public UnTreeDeviceParserPanel(
      anytype device,
      shared_ptr<UnTreeProcessContext> context = nullptr) : UnTreeDeviceParser(device,context){
    }

    public anytype encodeUserData(anytype userData) {
        // if more than 1 entry in dsUserData --> assume panel navigation
        if(getType(userData) == DYN_STRING_VAR){
          for(int i=2;i<=dynlen(userData);i++) {
              dyn_string splitted = strsplit(userData[i], c_panel_navigation_delimiter);
              if(dynlen(splitted)>1 && splitted[2] != ""){
                  splitted[2] = this.encodeDeviceName(splitted[2]);
                  _context.addDevice(UnTreeProcessorParam::MISSING_CANDIDATES,splitted[2]);
              }
              userData[i]=strjoin(splitted,c_panel_navigation_delimiter);
          }
        }
        return userData;
    }

    public anytype decodeUserData(mixed userData) {

        if(getType(userData) == STRING_VAR){
          strreplace(userData,"\r\r\n","\r\n");
          userData = strsplit(userData, "~");
        }
        if(getType(userData) == DYN_STRING_VAR){
          dyn_dyn_string temp;
          // if more than 1 entry in dsSplit --> assume panel navigation
          for(int i=1; i<dynlen(userData);++i){ //only if bigger than one element, thus starting from 1
            dyn_string splitted = strsplit(userData.at(i), c_panel_navigation_delimiter);
            while(dynlen(splitted) <3)
              dynAppend(splitted, "");
            temp.append(strsplit(splitted));
          }
          userData = temp;
        }
        if(getType(userData) == DYN_DYN_STRING_VAR){
          for(int i=2; i<=dynlen(userData); ++i){ //only if bigger than one element, thus starting from 1
            if(dynlen(userData[i]) < 3){ continue; }
            userData[i][2] = _context.decodeDeviceName(userData[i][2]);
          }
        }
        return userData;
    }

    // public mapping toMap(){
    //   reutrn makeMapping(
    //     UN_WINDOWTREE_PANEL,
    //     makeMapping(
    //       "panelname",unPanel_convertDpPanelNameToPanelFileName(this.getDeviceDp())
    //     )
    //   );
    // }

    public void exportToFile(shared_ptr<UnTreeFileParser> fileParser){
      this.isDeviceOk(UN_WINDOWTREE_PANEL); // would throw expection
      const string device = this.getDeviceDp();
      _context.addDevice(UnTreeProcessorParam::EXPORTED_DEVICES,device);
      fileParser.addNode(UN_WINDOWTREE_PANEL);
      fileParser.addValue("panelname",unPanel_convertDpPanelNameToPanelFileName(device));
      exportNavigationDevices(fileParser);
    }

    protected string getPanelDp(){
        return unPanel_convertPanelFileNameToDp(_data[NODE_PARAMNAME_PANELNAME]);
    }

    protected void _assertData(){
        FwException::assertNotNull(_data,"_data is empty");
        this._assertHasKey(NODE_PARAMNAME_PANELNAME);
        string panelDp = this.getPanelDp();
        panelDp.replace("-","");
        FwException::assert(panelDp!="","Panel name is empty");
        FwException::assert(unConfigGenericFunctions_nameCheck(panelDp), "Panel DP not correct:" + panelDp);
        if(dpExists(panelDp)) {
          FwException::assert(dpTypeName(panelDp)==UN_WINDOWTREE_PANEL,"Wrong panel dp type: " + dpTypeName(panelDp));
        }

        // check that the tree node device name and panel file name are the same
		// if(bTreeNodeDevice) {
		// 	sValue = unExportTree_removeSystemName(sTreeNodeDeviceName);
		// 	if(sValue != sDpName)
		// 		fwException_raise(exInfo, "ERROR", "unImportTree__UnPanel_checkTagSectionDataXml: tree device node and panel file name are different, "+sValue+", "+sDpName, "");
		// }
    }

    public string writeToDp(){
      this._assertData();
      string panelDp = _context.decodeDeviceName(this.getPanelDp());
      if(!dpExists(panelDp)){
        FwException::assert(dpCreate(panelDp, UN_WINDOWTREE_PANEL) == 0, "Could not create dp: " + panelDp);
        FwException::checkLastError("Could not create dp: " + panelDp);
      }
      this.setNavigation(panelDp);
      return panelDp;
    }

    // public bool matchesWithDP(){
    //     if(!UnTreeDeviceParser::matchesWithDP()){ return false; }
    //     string panelDp = this.getPanelDp();
    //     panelDp.replace("-","");
    //     if(!dpExists(panelDp)) { return false; }
    //     if(dpTypeName(panelDp) != UN_WINDOWTREE_PANEL) { return false; }
    //     return this._navigationMatches(panelDp);
    // }


};
