/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/fwTree/FwTree.ctl"
#uses "classes/unTree/deviceParsers/unTreeDeviceParser.ctl"

class UnTreeDeviceParserPlot : UnTreeDeviceParser {
    public UnTreeDeviceParserPlot(
      anytype device,
      shared_ptr<UnTreeProcessContext> context = nullptr) : UnTreeDeviceParser(device,context){
        if(dynlen(_tags) == 0){
          _tags = fwTrending_getPlotTreeTagNames();
          for ( int idx=1 ; idx<=dynlen(_tags) ; ++idx ){
            _tags[idx] = substr(_tags[idx],1);
          }
        }
    }

    static protected dyn_string _tags;

    // public string getCustomUserData() {
    //     return "";
    // }

    static void adjustToEightElements( dyn_dyn_string &table , int entry , string value , bool incre = false ) {
      const int len = dynlen(table[entry]);
      if(len == 8) { return; }
      for ( int i=len ; i<8 ; ++i) {
        if (!incre){
          dynAppend ( table[entry] , value );
        }
        else {
          string datas;
          sprintf(datas,value,i+1);
          dynAppend ( table[entry] , datas );
        }
      }
      for (int i=len; i>8; --i){
        dynRemove ( table[entry] , i );
      }
    }

    protected void getPlotData(dyn_dyn_string &ddsPlotData, int &itemslen){
      dyn_string exInfo;
      fwTrending_getPlot(this.getDeviceDp(), ddsPlotData, exInfo);
      FwException::checkRaise(exInfo);
      const int plotlen = dynlen(_tags);
      const int trendlen = dynlen(ddsPlotData);
      itemslen = ( trendlen < plotlen ? trendlen : plotlen );

      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_DPES_X,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_DPES,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_LEGENDS_X,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_LEGENDS,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_COLORS,"FwtrendingCurve%d",TRUE);
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_AXII_X,"FALSE");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_AXII,"FALSE");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_CURVES_HIDDEN,"FALSE");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_RANGES_MIN_X,"0");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_RANGES_MAX_X,"0");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_RANGES_MIN,"0");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_RANGES_MAX,"0");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_CURVE_TYPES,"1");

      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_AXII_POS,"0");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_AXII_LINK,"0");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_AXII_X_FORMAT,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_AXII_Y_FORMAT,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_LEGEND_VALUES_FORMAT,"");
      adjustToEightElements(ddsPlotData,fwTrending_PLOT_OBJECT_ALARM_LIMITS_SHOW,"FALSE");
    }

    public void exportToFile(shared_ptr<UnTreeFileParser> fileParser){
        const string device = this.getDeviceDp();
        this.isDeviceOk(UN_TRENDTREE_PLOT); // would throw expection
        _context.addDevice(UnTreeProcessorParam::EXPORTED_DEVICES,device);

        anytype topNode = fileParser.addNode(UN_TRENDTREE_PLOT);
        fileParser.addValue("dpname",_context.encodeDeviceName(device));

        dyn_dyn_string ddsPlotData;
        mapping exportPlotData;
        int itemslen;
        this.getPlotData(ddsPlotData,itemslen);

        for ( int idx = 1 ; idx <= itemslen ; ++idx ) {
          if ( ! fwTrending_isPlotTreeTagExportable(idx) ){
            continue;
          }
          exportPlotData[_tags[idx]] = makeDynMixed();

          for ( int cnt = 1 ; cnt <= dynlen(ddsPlotData[idx]) ; ++cnt ){
            string iPlotData = ddsPlotData[idx][cnt];
            if ( idx == fwTrending_PLOT_OBJECT_DPES || idx == fwTrending_PLOT_OBJECT_LEGENDS) {
              iPlotData = _context.encodeDeviceName(iPlotData);
            }
            exportPlotData[_tags[idx]].append(iPlotData);
          }
        }
        fileParser.addValue("plotconfiguration",exportPlotData,-1,"");
        fileParser.setCurrentTop(topNode);
        exportNavigationDevices(fileParser); 
    }
};
