/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/unTree/unTreeProcessContext.ctl"
#uses "classes/fwStdUi/FwStdUi.ctl"
#uses "classes/fwTree/FwTree.ctl"

class UnTreeFileParser {
    protected shared_ptr<UnTreeProcessContext> _context;
    protected string _path;
    protected mixed _currentTop;
    protected anytype _rootNode;

    public anytype getTopNode(){ return _rootNode; }
    public anytype getCurrentTop(){ return _currentTop; }

    public void setCurrentTop(anytype nodeId){
        _currentTop = nodeId;
    }

    public bool checkFileAccessImport(const string fileName, string &details){
        if(!isfile(fileName)){
            if(isdir(fileName)){
                details = "The path is directory: " + fileName;
                return false;
            } 
            details = "File does not exist:" + fileName;
            return false;
        }
        if( access(fileName, R_OK) == 0 ){
            details = "File is OK for read." + fileName;
            return true;
        } else {
            details = "File exist but cannot be read: " + fileName;
            return false;
        }
    }

    public bool checkFileAccessExport(const string fileName, string &details){
        if(!isfile(fileName)){
            if(isdir(fileName)){
                details = "The path is directory: " + fileName;
                return false;
            } else {
                details = "New file will be created";
                return true;
            }
        }
        if( access(fileName, F_OK) == 0 ){ // file exists
            if(_context.get(UnTreeProcessorParam::OVERWRITE_OK)){
                details = "File will be overwriten: " + fileName;
            }
            if(!_context.get(UnTreeProcessorParam::SILENT_MODE)){
                if(!FwStdUi::dialogConfirm("Are you sure you want to overwrite the file?\n" + fileName)){
                    details = "Aborted by user";
                    return false;
                } else {
                    details = "File will be overwriten: " + fileName;
                }
            }
        }
        if( access(fileName, W_OK) == 0 ){ // file writable
            if(details == ""){
                details = "New file will be created: " + fileName;
            }
            return true;
        } else {
            details = "File not accessible for writing: " + fileName;
            return false;
        }
    }

    public UnTreeFileParser(shared_ptr<UnTreeProcessContext> context = nullptr){
        if(equalPtr(context,nullptr)){
            _context = new UnTreeProcessContext();
        } else {
            _context = context;
        }
    }

    public initializeExportFile(const string filePath, const FwTreeNodePtr node){}
    public initializeImportFile(const string filePath){}
    public mixed addNode(const string key, mixed parent=-1, const bool setAsTop = true){ return nullptr; }
    public void addValue(const string key, const anytype value, mixed parent = -1, const string sep = UN_PARAMETER_DELIMITER){}
    public bool writeToFile(){ return false; }
    public void resetTop(anytype topNode = -1) {}
    public void removeIfEmpty(const mixed nodeId){}
    public void parseFromFile(const string filePath){}
    public shared_ptr<mapping> asMap() { return _rootNode; }

    public static dyn_string decodeList(mixed input, const string sep = UN_PARAMETER_DELIMITER, const int size = -1){
        if(getType(input) == DYN_STRING_VAR){
            while(dynlen(input) < size){ input.append(""); }
            return input;
        }
        if(getType(input) == STRING_VAR){
            dyn_string result = strsplit(input,sep);
            while(dynlen(result) < size){ result.append(""); }
            return result;
        }
        DebugTN(__FILE__,__LINE__,"unsupported input:",input);
        FwException::raise("Unsupported list format: " + getTypeName(input));
    }
};