/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/unTree/fileParsers/unTreeFileParser.ctl"

class UnTreeFileParserJson:UnTreeFileParser{

    public UnTreeFileParserJson(shared_ptr<UnTreeProcessContext> context = nullptr):UnTreeFileParser(context){
        _rootNode = new mapping();
        _currentTop = _rootNode;
    }

    public void parseFromFile(const string filePath){
        _rootNode = jsonDecode(filePath);
        _currentTop = _rootNode;
    }

    public initializeExportFile(const string filePath, const FwTreeNodePtr node){
        _path = filePath;
        FwTreeNodePtr rootNode = node.isRoot() ? node : node.getRootNode();
        _rootNode.clear();
        _currentTop = _rootNode;
        this.addNode("tree",new mapping());
        this.addValue("version",_context.get(UnTreeProcessorParam::FORMAT_VERSION));
        const string xmlName = _context.checkValue(UnTreeProcessorParam::TREE_NAME,"",rootNode.getName());
        // unExportTree_uniqueTreeTagNames();
        this.addValue("name",xmlName);
        _rootNode.insert("_FwTreeNode",makeDynMixed());
    }

    protected void insert(mixed target, mixed key, mixed node){
        if(getType(target)==INT_VAR && target == -1){
            target = _currentTop;
        }
        if(getType(target) == DYN_MIXED_VAR){
            target.append(node);
        } else {
            target.insert(key,node);
        }
    }

    public mixed addNode(const string key, mixed parent=-1, const bool setAsTop = true){
        if(key=="otherdevice") {
            if(_rootNode.contains(key)){
                return _rootNode[key];
            }
            shared_ptr<dyn_mixed> otherDevs = new dyn_mixed();
            _rootNode.insert("otherdevice",otherDevs);
            return otherDevs;
        } 
        shared_ptr<mapping> newNode = new mapping();
        if(key == "_FwTreeNode"){
            // shared_ptr<dyn_mixed> nodes = _rootNode["_FwTreeNode"];;
            // nodes.append(newNode);
            // dynAppend(_rootNode["_FwTreeNode"],newNode);
            _rootNode["_FwTreeNode"].append(newNode);
        } else {
            this.insert(parent,key,newNode);
        }
        if(setAsTop) { _currentTop = nullptr; _currentTop = newNode; }
        return newNode;
    }

    public void resetTop(anytype topNode = -1) {
        if(getType(topNode)==INT_VAR && topNode == -1){
            _currentTop = _rootNode;
        } else {
            _currentTop = topNode;
        }
    }

    public void addValue(const string key, const anytype value, mixed parent = -1, const string sep = UN_PARAMETER_DELIMITER){
        if(value == "" && _context.get(UnTreeProcessorParam::SPARSE)){
            return;
        }
        this.insert(parent,key,value);
    }

    public bool writeToFile(){
        file f=fopen(_path,"w");
        fputs(jsonEncode(_rootNode,!_context.get(UnTreeProcessorParam::BEAUTIFY)),f);
        fclose(f);
        return true;
    }

    public void removeIfEmpty(const mixed nodeId){
        if(getType(nodeId) == DYN_MIXED_VAR){
            if(dynlen(nodeId) != 0){
                return;
            }
        } else if(getType(nodeId) == MAPPING_VAR){
            if(dynlen(mappingKeys(nodeId)) != 0){
                return;
            }
        } else {
            return;
        }

        if(getType(_currentTop) == MAPPING_VAR){
            _currentTop.remove(nodeId);
            if(_currentTop == nodeId){
                this.resetTop();
            }
        }
    }
};