/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "CtrlXml"
#uses "unExportTree.ctl"
#uses "classes/fwTree/FwTree.ctl"
#uses "unExportTree.ctl"
#uses "classes/unTree/deviceParsers/unTreeDeviceParserFactory.ctl"
#uses "classes/unTree/deviceParsers/unTreeNodeParser.ctl"
#uses "classes/unTree/fileParsers/unTreeFileParserFactory.ctl"
// #uses "classes/fwTree/FwTree_RepositoryImpl.ctl"
#uses "classes/unTree/unTreeRepositoryImpl.ctl"
#uses "classes/unTree/delegates/unTreeDelegateFactory.ctl"

// top level of export, responsible for an export file location and unique access
class UnTreeProcessor{
  #event sendLogMessage(string msg, string type = "INFO")

  // protected static shared_ptr<UnTreeProcessor> _theInstance = nullptr;

  protected shared_ptr<UnTreeProcessContext> _context;
  protected shared_ptr<UnTreeFileParser> _fileParser;
  protected static bool _busy = false;
  protected bool _stop = false;
  protected bool _hasWarnings = false;
  protected string _action = "";
  protected shared_ptr< UnTree_RepositoryReadImpl > _readRepo = nullptr;

  // protected dyn_string _nodesList;

  public UnTreeProcessor(){ // just to do singleton
      _context = new UnTreeProcessContext();
  }

  public void setParams(shared_ptr<UnTreeProcessContext> context){
    _context = nullptr;
    _context = context;
  }

  public shared_ptr<UnTreeProcessContext> contextPtr(){
    return _context;
  }

  public bool hasWarnings(){
    return _hasWarnings;
  }


  protected void getTreeInfo(dyn_string &modes, string &nodeTypes, string &nodeNames, int &refresh){
    const string dp = "fwTT_" + UnTreeDelegateFactory::getTreeTypeDP(_context.get(UnTreeProcessorParam::TREE_TYPE),TREE_WIDGET_PREVIEW);
    FwException::assertDP(dp,"_FwTreeType");
    if(dpGet(dp+".nodeTypes",nodeTypes,dp+".nodeNames",nodeNames,dp+".modes",modes,dp+".refresh",refresh) != 0 ) {
      FwException::raise("Could not get tree dpes: " + dp);
    }
    FwException::assert(dynlen(getLastError()) == 0,"Could not get tree dpes: " + dp);
  }

  protected void setTreeInfo(dyn_string modes, string nodeTypes, string nodeNames, int refresh){
    const string dp = "fwTT_" + UnTreeDelegateFactory::getTreeTypeDP(_context.get(UnTreeProcessorParam::TREE_TYPE),TREE_WIDGET_PREVIEW);
    FwException::assertDP(dp,"_FwTreeType");
    if(dpSetWait(dp+".nodeTypes",nodeTypes,dp+".nodeNames",nodeNames,dp+".modes",modes,dp+".refresh",refresh) != 0 ) {
      FwException::raise("Could not set tree dpes: " + dp);
    }
    FwException::assert(dynlen(getLastError()) == 0,"Could not set tree dpes: " + dp);
  }

  public mapping getNodesStatuses(){
    return _context.get(UnTreeProcessorParam::NODES_STATUSES);
  }

  public bool isBusy(){ return _busy; }
  public string getActionDetails(){
    return "Action: " + _action + " tree type: " + _context.get(UnTreeProcessorParam::TREE_TYPE);
  }

  // protected void reloadNodesList(const FwTreeNodePtr node){
  //   _nodesList.append(node.getDP());
  //   FwTreeNodePVec children = node.getChildren();
  //   for ( int i=0 ; i<children.count() ; ++i ) {
  //     this.reloadNodesList(children.at(i));
  //   }
  // }

  public void processAction(const string action, const FwTreeNodePtr sourceNode, const string fileName, mapping context, const FwTreeNodePtr targetNode = nullptr){
    if(action == "stop"){
      _stop = true;
      return;
    }
    if(_busy){
      return;
    }
    _busy = true;
    _action = action;
    _context.fromMapping(context);
    // _nodesList.clear();
    // this.reloadNodesList(sourceNode);
    try{
      switch(action){
        case "read":
          this.fromFile(sourceNode,fileName);
          break;
        case "import":
          this.import(sourceNode,targetNode);
          break;
        case "export":{
          if(context.value(UnTreeProcessorParam::TREE_ROOT,"") == ""){
            context[UnTreeProcessorParam::TREE_ROOT] = context[UnTreeProcessorParam::TREE_TYPE] + "Tree";
          }
          _context.fromMapping(context);
          this.toFile(sourceNode,fileName);
          break;
        }
        default:
          FwException::raise("Unsupported action: " + action);
      }
    } catch {
      _busy = false;
      FwException::rethrow();
    }
    _busy = false;
  }

  protected void checkStop(){
    // delay(0,_context.get(UnTreeProcessorParam::DELAY),0); // adding delay for debugging purposes
    if(!_stop){ return; }
    _stop = false;
    _busy = false;
    FwException::raise("Aborted by user");
  }

// unExportTree_exportTreeDataXml(int _xmlDoc, int _xmlTop, string dpTree, string sNodeName, bool bKeepLocalSystem, string sSystemNameToReplace, bool noSilent, dyn_string &exInfo)
  protected void toFile(const FwTreeNodePtr node, const string fileName){
    if(node == nullptr) { // || (node.getChildren().count() == 0)){
      FwException::raise("Could not export empty node - processor");
    }

    _fileParser = nullptr;
    _context.set(UnTreeProcessorParam::FORMAT, fileName.split(".").last());
    _fileParser = UnTreeFileParserFactory::getParser(_context);
    string details;
    if(!_fileParser.checkFileAccessExport(fileName, details)){
      FwException::raise("Could not write to file - " + details);
    }
    sendLogMessage(details);
    this.checkStop();
    _fileParser.initializeExportFile(fileName,node);

    string names, types;
    dyn_string modes;
    int refresh;
	try{
		this.getTreeInfo(modes, types, names, refresh);
	} catch {
      _context.rethrowIfNeeded();
	}

    this.checkStop();
    string root;
    _context.asMap().value(UnTreeProcessorParam::TREE_ROOT,"");
    if(root == "") root = node.getRootNode().getName();
-    // unExportTree_uniqueTreeTagNames();

    _fileParser.addValue( "treeparameters", makeDynString(root,types,names,refresh) );
    _fileParser.addValue( "treemodes", modes);

    sendLogMessage("Exporting from top-level tree node data: " + fileName);
    this.exportNode(node,true);
    this.exportMissingDevices(node);

    sendLogMessage("Writing to file: " + fileName);
    _fileParser.writeToFile();
  }

// unExportTree_exportSubTreeDataXml(_xmlDoc, int _xmlTop, string sNodeName, bool bKeepLocalSystem, string sSystemNameToReplace, bool noSilent, dyn_string &exInfo)
protected void exportNode(const FwTreeNodePtr node, const bool rootNode = false){
  if(node == nullptr){
    _context.throwIfNeeded("Could not export empty node");
    return;
  }
  this.checkStop();

  if(!rootNode){
    shared_ptr<UnTreeNodeParser> nodeParser = new UnTreeNodeParser(node, _context);
    sendLogMessage("Exporting node: " + node.getName());
    try{
      nodeParser.exportToFile(_fileParser);
      _context.setStatus(node.getDP(),NODE_STATUS_OK);
    } catch {
      _context.rethrowIfNeeded();
      _context.setStatus(node.getDP(),NODE_STATUS_ERR);
      return;
    }
  }

  FwTreeNodePVec children = node.getChildren();
  for ( int i=0 ; i<children.count() ; ++i ) {
    this.exportNode(children.at(i));
  }
}

protected void exportMissingDevices(const FwTreeNodePtr node){
    const dyn_string missingCandidates = _context.get(UnTreeProcessorParam::MISSING_CANDIDATES);
    _context.set(UnTreeProcessorParam::MISSING_CANDIDATES,makeDynString());
    const int missingSize = dynlen(missingCandidates);
    if(missingSize == 0){
      return;
    }
    sendLogMessage("Exporting the other missing devices data... ");
    const dyn_string exportedDevices = _context.get(UnTreeProcessorParam::EXPORTED_DEVICES);
    const anytype otherDeviceNode = _fileParser.addNode("otherdevice",_fileParser.getTopNode());
    for(int i=0; i<missingSize; ++i){
      this.checkStop();
      if(!dynContains(exportedDevices, missingCandidates.at(i))){
        // FwTreeNodePtr missingDev = FwTree_Repository::getInstance().get(missingCandidates.at(i));
        try{
          shared_ptr<UnTreeDeviceParser> deviceParser = UnTreeDeviceParserFactory::createDeviceParser(missingCandidates.at(i),_context);
          deviceParser.exportToFile(_fileParser);
          _fileParser.setCurrentTop(otherDeviceNode);
        } catch {
          _context.rethrowIfNeeded();
        }
      }
    }
    _fileParser.removeIfEmpty(otherDeviceNode);
}

protected void merge(FwTreeNodePtr mergeFrom, FwTreeNodePtr mergeTo, bool subTreeOnly=false){
  FwException::assertNotNull(mergeFrom,"Could not merge from empty source");
  FwException::assertNotNull(mergeTo,"Could not merge into empty target");
    FwTreeNodePVec newChildren=mergeFrom.getChildren();
    FwTreeNodePVec curChildren=mergeTo.getChildren();
    if (mergeFrom.equals(mergeTo,true,false)) return;


}

// protected void areNodesSame(FwTreeNodePtr mergeFrom, FwTreeNodePtr mergeTo){
//   return false;
// }

// protected void merge(FwTreeNodePtr mergeFrom, FwTreeNodePtr mergeTo, bool subTreeOnly=false)
// {
//     FwException::assert(equalPtr(mergeTo._repo, this._selfPtr),"Cannot merge trees: wrong repo in the target tree, "+mergeTo._dp);

//     // let's firstly process only the children...
//     FwTreeNodePVec newChildren=mergeFrom.getChildren();
//     FwTreeNodePVec curChildren=mergeTo.getChildren();
//     FwTreeNodePVec updatedChildren;
//     FwTreeNodePVec removedChildren;


//     // diff, including the list of children checked recursively yet with no order-checking
//     if (mergeFrom.equals(mergeTo,true,false)) return;

//     // if (!subTreeOnly) mergeTo.assign(mergeFrom); // assign all properties but not links...

//     mapping curChildrenMap;
//     for (int i=0;i<curChildren.count();i++) {
//       curChildrenMap[curChildren.at(i).getName()] = curChildren.at(i);
//     }

//     for (int i=0;i<newChildren.count();i++) {
//       FwTreeNodePtr newNode = newChildren.at(i);
//       if( curChildrenMap.contains(newNode.getName()) ){
//         FwTreeNodePtr oldNode = curChildrenMap[newNode.getName()];
//         if(!areNodesSame(oldNode,newNode)){
//           string type, linked;
//           bool master;
//           dyn_string userData;
//           oldNode.setUserData(newNode.getUserData)
//     public void       getProps(string &aNodeType, string &aLinkedObject, bool &aIsMaster, dyn_string &aUserData)
//           oldNode.set
//         }
//         updatedChildren.append(oldNode);
//       } else {

//       }
//     }

//     for (int i=0;i<curChildren.count();i++) {

//     // now treat the list of children, with recursion

//     for (int i=0;i<newChildren.count();i++) {
//         FwTreeNodePtr c = newChildren.at(i);
//         // now we check by node *name* if a similar one exists
//         FwTreeNodePtr newChild=nullptr;
//         vector<int> foundIdxs = curChildren.indexListOf("name",c.name);
//         if (foundIdxs.isEmpty()) {
//             newChild = create(c.name,mergeTo,c.type, c.linkedObj, c.userData, c.isMaster);
//         } else {
//             newChild=curChildren.at(foundIdxs.first());
//         }
//         // recurse the merge over children...
//         merge(c,newChild);
//     }
//     persist(mergeTo);
// }

protected void duplicateNode(FwTreeNodePtr source, FwTreeNodePtr target, const bool assignParent = true){
    FwTreeNodePVec kids=source.getChildren();
    for (int i=0;i<kids.count();i++) {
        FwTreeNodePtr c = kids.at(i);
        FwTreeNodePtr newChild=nullptr;
        newChild = target.getRepo().create(c.name,target,c.type, c.linkedObj, c.userData, c.isMaster);
        duplicateNode(c,newChild);
    }
    if(assignParent)
      target.assign(source);

    shared_ptr<UnTreeDeviceParser> devParser = _readRepo.getDeviceParser(target.getLinkedObj());
    if(devParser != nullptr){
      // nodeParser = new UnTreeNodeParser(target, _context);
      devParser.writeToDp();
    }
}

public static void mergeNodes(FwTreeNodePtr mergeFrom, FwTreeNodePtr mergeTo)
{
    // FwException::assert(equalPtr(mergeTo._repo, this._selfPtr),"Cannot merge trees: wrong repo in the target tree, "+mergeTo._dp);

    // let's firstly process only the children...
    FwTreeNodePVec newChildren=mergeFrom.getChildren();
    FwTreeNodePVec curChildren=mergeTo.getChildren();

    // diff, including the list of children checked recursively yet with no order-checking
    if (mergeFrom.equals(mergeTo,true,false)) return;

    // if (!subTreeOnly) mergeTo.assign(mergeFrom); // assign all properties but not links...

    // now treat the list of children, with recursion

    for (int i=0;i<newChildren.count();i++) {
        FwTreeNodePtr c = newChildren.at(i);
        // DebugN("Processing kid: ", c.name);
        // now we check by node *name* if a similar one exists
        FwTreeNodePtr newChild=nullptr;
        vector<int> foundIdxs = curChildren.indexListOf("name",c.name);
        if (foundIdxs.isEmpty()) {
            newChild = mergeTo.getRepo().create(c.name,mergeTo,c.type, c.linkedObj, c.userData, c.isMaster);
            // mergeTo.getRepo().clone(c,mergeTo,true);
            // newChild = mergeTo.getRepo().create();
        } else {
            newChild=curChildren.at(foundIdxs.first());
        }
        // recurse the merge over children...
        mergeNodes(c,newChild);
    }
    mergeTo.assign(mergeFrom);
    // mergeTo.getRepo().persist(mergeTo);
}

protected void import(const FwTreeNodePtr sourceNode, FwTreeNodePtr targetNode){
  if(targetNode == nullptr){
    const string treeName = "fwTN_" + _context.get(UnTreeProcessorParam::TREE_TYPE) + "Tree";
    targetNode = FwTree_Repository::getDefault().get(treeName);//.getRootNode();
    FwException::assertNotNull(targetNode, "Could not find default top node for import!");
  }
    FwTreeNodePVec nodes = targetNode.getChildren();
    for(int i=0; i<nodes.count(); ++i){
        nodes[i].removeMe(true);
    }
    duplicateNode(sourceNode,targetNode,false);
}

protected void fromFile(const FwTreeNodePtr node, const string fileName){
    if(node == nullptr){
      FwException::raise("Could not import into non-existent node");
    }

    this.checkStop();

    _fileParser = nullptr;
    _context.set(UnTreeProcessorParam::FORMAT, fileName.split(".").last());
    _fileParser = UnTreeFileParserFactory::getParser(_context);
    string details;
    if(!_fileParser.checkFileAccessExport(fileName, details)){
      FwException::raise("Could not read from file - " + details);
    }
    sendLogMessage(details);
    sendLogMessage("Starting tree file parsing: " + fileName);
    _fileParser.parseFromFile(fileName);
    _readRepo = node.getRepo();
    this.readTree(node);

    // if(isFunctionDefined("unMessageText_sendException")) {
    //   dynClear(exMessage);
    //   fwException_raise(exMessage, "INFO", "Importing Tree data => " + sFileName, "");
    //   unMessageText_sendException("*", "*", "ImportTree", "user", "*", exMessage);
    // }

    // unImportTree_checkTreeDataXml(xmlDoc, xmlEls, sTreeName, bKeepSystemName, exInfoTemp);

    sendLogMessage("Read done");
}

public static void clearTree(FwTreeNodePtr tree){
  FwTreeNodePVec toRemove = tree.getRootNode().getChildren();// tree.getRootNode().getClipboard().getChildren();
 // dynAppend(toRemove,tree.getRootNode().getChildren());
  for(int i=0; i<toRemove.count(); ++i){
      toRemove[i].removeMe(true);
  }
}

public static void draw(const FwTreeNodePtr tn, const string title = "", const string indent = ""){
  if(tn==nullptr) return;
  if(title != "") DebugN(title);
  DebugN(indent + tn.getName());
  FwTreeNodePVec kids = tn.getChildren();
  for(int i=0;i<kids.count();++i){
    draw(kids.at(i),"",indent + "-");
  }
}

protected void readTree(const FwTreeNodePtr parent){
  this.clearTree(parent);
  sendLogMessage("parsing tree fromat to mapping...");
  shared_ptr<mapping> asMap = _fileParser.asMap();
  dyn_string treeParams = asMap.value("treeparameters",makeDynString());
  while(dynlen(treeParams)<4){ treeParams.append(""); }
  // this.setTreeInfo(
  //   _context,
  //   asMap.value("treemodes",makeDynString()),
  //   treeParams.at(1),
  //   treeParams.at(2),
  //   (int)treeParams.at(3)
  // );
  this.checkStop();
  _context.set(UnTreeProcessorParam::TREE_ROOT,treeParams.at(0));
  if(
    asMap.contains("_FwTreeNode") && 
    getType(asMap["_FwTreeNode"]) == MAPPING_VAR
  ){
    parent.getRepo().readNode(parent,asMap["_FwTreeNode"]);
  } else {
    const int nodesNum = dynlen(asMap.value("_FwTreeNode",makeDynMixed()));
    vector< shared_ptr<UnTreeNodeParser> > nodes;
    for(int i=0; i<nodesNum; ++i){
      this.checkStop();
      parent.getRepo().readNode(parent,asMap["_FwTreeNode"].at(i));
    }
  }
  this.checkStop();
  parent.getRepo().applyHierarhy();
  this.checkStop();
  parent.getRepo().fixChildrenOrder();
}

};
