/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/fwTree/FwTree_SimpleImpl.ctl"
#uses "classes/unTree/unTreeProcessContext.ctl"
#uses "classes/unTree/deviceParsers/unTreeDeviceParser.ctl"
#uses "classes/unTree/deviceParsers/unTreeNodeParser.ctl"

class UnTree_RepositorySimple: FwTree_RepositorySimple {
    protected UnTree_RepositorySimple(){} // disable to make sure we only get share'able instances, and we have the _selfPtr right

    // main factory method to get new instances. Note: NOT A SINGLETON!
    public static shared_ptr<FwTree_Repository> getInstance()
    {
        shared_ptr<FwTree_Repository> newInstance = new UnTree_RepositorySimple;
        newInstance._selfPtr=newInstance;

        return newInstance;
    }
    protected string generateId(string name)
    {
        if(!treeNodeMap.contains(name)){
            return name;
        }
        string id;
        sprintf(id,"unTN_&%04d%s",_cnt++,name);
        return id;
    }
};

class UnTree_RepositoryReadImpl: UnTree_RepositorySimple {

    // private static int _cnt=0; // to generate instance ids...
    // protected mapping treeNodeMap; // key: id, value: shared_ptr<FwTreeNode>, for quick lookups
    // protected FwTreeNodePVec allTreeNodes;
    // protected FwTree_RepositorySimple(){} // disable to make sure we only get share'able instances, and we have the _selfPtr right
    protected mapping _parsers; // nodeParsers for all nodes
    protected mapping _devicesParsers; // nodeParsers for all nodes
    protected shared_ptr<UnTreeProcessContext> _context = nullptr;

    protected UnTree_RepositoryReadImpl(){}

    public dyn_string getDevices(){
        dyn_string result;
        for(int i=0; i<_devicesParsers.count(); ++i){
            result.append(_devicesParsers.keyAt(i));
        }
        return result;
    }

    // main factory method to get new instances. Note: NOT A SINGLETON!
    public static shared_ptr<FwTree_Repository> getInstance()
    {
        shared_ptr<FwTree_Repository> newInstance = new UnTree_RepositoryReadImpl;
        newInstance._selfPtr=newInstance;

        return newInstance;
    }

    // creates a new node...
    // public FwTreeNodePtr createRootNode(string name)

    // another factory, allowing to also specify the nodeId, eg. when reading from a file
    // public FwTreeNodePtr createNodeWithId(string nodeId, string name, FwTreeNodePtr parent)
    // public FwTreeNodePtr create(string name, FwTreeNodePtr parent,
    // public FwTreeNodePtr get(string treeNodeId, bool allowInvalid=false)
    // public FwTreeNodePVec findBy(string memberName, mixed value, ...)
    // public void drop(FwTreeNodePtr tn, bool recursively=false)
    // public void persist(FwTreeNodePtr tn)
    // public void renameNode(FwTreeNodePtr tn, string newNodeName)

    public shared_ptr<UnTreeDeviceParser> getDeviceParser(const string dp){
        return _devicesParsers.value(dp,nullptr);
    }

    public void setContext(shared_ptr<UnTreeProcessContext> context){
        assignPtr(this._context,context);
    }

    public void readNode(const FwTreeNodePtr parent, const shared_ptr<mapping> nodeData){
        shared_ptr<UnTreeNodeParser> nodeParser = new UnTreeNodeParser(nullptr,_context);
        nodeParser.createFromMap(parent, nodeData);
        _parsers[nodeParser.getNode().getDP()] = nodeParser;
        shared_ptr<UnTreeDeviceParser> dev = nodeParser.getDeviceParser();
        if(dev!=nullptr)
            _devicesParsers[dev.getDeviceDp()] = dev;
    }

    protected shared_ptr<UnTreeNodeParser> _getParser(const int idx){
        FwException::assert(idx<allTreeNodes.count(), "Paser index excedeed: "+idx);
        const string id = allTreeNodes.at(idx).getDP();
        const bool found = _parsers.contains(id);
        if(!found){
            _context.throwIfNeeded("Parser not found: " + id + strjoin(mappingKeys(_parsers,",")));
            return nullptr;
        }
        return _parsers[id];
    }

    public void applyHierarhy(){
        for(int i=1; i<allTreeNodes.count(); ++i){ //skipping first, root node
            shared_ptr<UnTreeNodeParser> parser = this._getParser(i);
            if(!equalPtr(parser,nullptr)){
                parser.applyHierarhy();
            }
        }
    }

    public void fixChildrenOrder(){
        for(int i=1; i<allTreeNodes.count(); ++i){ //skipping first, root node
            shared_ptr<UnTreeNodeParser> parser = this._getParser(i);
            if(!equalPtr(parser,nullptr)){
                parser.fixChildrenOrder();
            }
        }
   }
};
